# Plugin for TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) Evolved Media Network 2005
# Copyright (C) Spanlink Communications 2006
# and TWiki Contributors. All Rights Reserved. TWiki Contributors
# are listed in the AUTHORS file in the root of this distribution.
# NOTE: Please extend that file, not this notice.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# For licensing info read LICENSE file in the TWiki root.
#
# Author: Crawford Currie http://c-dot.co.uk
#
# This plugin helps with permissions management by displaying the web
# permissions in a big table that can easily be edited. It updates
# WebPreferences in each affected web.
package TWiki::Plugins::WebPermissionsPlugin::Core;

use strict;
use Switch;
use constant DEBUG => 0; # toggle me
# BUGO: because WEBPERMISSIONS is using the same view to change and display,
# updated ACL's do not apply to this view.
# TWiki has already loaded the permissions that it is using, and some
# random plugins have already processed things based on the ACLs prior to
# the users change. THIS IS HORRIGIBILE
# IMO it needs to be either rest or save, though i'm going to lean to rest
# especially as we don't want to continue to presume that ACLs are topic based
# the advantage with save, is that it will re-direct to view on success (and
# resolve permissions issues)
sub WEBPERMISSIONS {
    my( $session, $params, $topic, $web ) = @_;
    my $query = $session->{cgiQuery};
    my $action = $query->param( 'web_permissions_action' );
    my $editing = $action && $action eq 'Edit';
    my $saving =  $action && $action eq 'Save';

    my @modes = split(/[\s,]+/,
                      $TWiki::cfg{Plugins}{WebPermissionsPlugin}{modes} ||
                        'VIEW,CHANGE' );

    my @webs = TWiki::Func::getListOfWebs( 'user' );
    my $chosenWebs = $params->{webs} || $query->param('webs');
    if( $chosenWebs ) {
        @webs = _filterList($chosenWebs, @webs);
    }
	
    my @knownusers;
    my $chosenUsers = $params->{users} || $query->param('users');

    my %table;
    foreach $web ( @webs ) {
        #TODO: use TWiki::Func::getRegularExpression(webNameRegex)
        next unless ($web=~/^(\w*\/?\w*)$/); #untaint before we do anything
        $web = $1;
		#&TWiki::Func::writeDebug('untainted '.$web);
        my ( $acls, $inheritedPermissions ) = _getACLs( \@modes, $web );

        unless( scalar( @knownusers )) {
            @knownusers = keys %$acls;
            if ($chosenUsers) {
                @knownusers = _filterList($chosenUsers, @knownusers);
            }
        }

        if( $saving ) {
            my $changes = 0;
            foreach my $user ( @knownusers ) {
                foreach my $op ( @modes ) {
                    my $onoff = $query->param($user.':'.$web.':'.$op);
                    if( $onoff && !$acls->{$user}->{$op} ||
                          !$onoff && $acls->{$user}->{$op} ) {
                        $changes++;
                        $acls->{$user}->{$op} = $onoff;
                    }
                }
            }
            # Commit changes to ACLs
            if( $changes ) {
				my $context = 'WEB';
                my $settings = _parseCustomACLs( \@modes, $acls, $context );
				_writeACLs( \@modes, $settings, $web, $TWiki::cfg{WebPrefsTopicName}, $context );
            }
        }
        $table{$web} = $acls;
    }

    # Generate the table
    my $tab = '';

    my %images;
    foreach my $op ( @modes ) {
        if( -f TWiki::Func::getPubDir().'/TWiki/WebPermissionsPlugin/'.$op.'.gif' ) {
              $images{$op} =
                CGI::img( { src => TWiki::Func::getPubUrlPath().
                              '/TWiki/WebPermissionsPlugin/'.$op.'.gif' } );
              $tab .= $images{$op}.' '.$op;
        } else {
            $images{$op} = $op;
        }
    }

    $tab .= CGI::start_table( { border => 1, class => 'twikiTable' } );

    my $repeat_heads = $params->{repeatheads} || 0;
    my $repeater = 0;
    my $row;

    foreach my $user ( sort @knownusers ) {
        unless( $repeater ) {
            $row = CGI::th( '' );
            foreach $web ( @webs ) {
                $row .= CGI::th( $web );
            }
            $tab .= CGI::Tr( $row );
            $repeater = $repeat_heads;
        }
        $repeater--;
        $row = CGI::th( "$user " );
        foreach $web ( sort @webs ) {
            my $cell;
            foreach my $op ( @modes ) {
                if( $editing ) {
                    my %attrs = ( type => 'checkbox', name => $user.':'.$web.':'.$op );
                    $attrs{checked} = 'checked' if $table{$web}->{$user}->{$op};
                    $cell .= CGI::label( ($images{$op} || $op).CGI::input( \%attrs ));
                } elsif( $table{$web}->{$user}->{$op} ) {
                    $cell .= $images{$op} || $op;
                }
            }
            $row .= CGI::td( $cell );
        }
        $tab .= CGI::Tr( $row );
    }
    $tab .= CGI::end_table();

    if( $editing ) {
        $tab .= CGI::submit( -name => 'web_permissions_action', -value => 'Save',  -class => 'twikiSubmit');
        $tab .= CGI::submit( -name => 'web_permissions_action', -value => 'Cancel',  -class => 'twikiSubmit' );
    } else {
        $tab .= CGI::submit( -name => 'web_permissions_action', -value => 'Edit',  -class => 'twikiSubmit' );
    }
    my $page = CGI::start_form(
        -method => 'POST',
        -action => TWiki::Func::getScriptUrl( $web, $topic, 'view').
          '#webpermissions_matrix' );
    $page .= CGI::a({ name => 'webpermissions_matrix'});
    if( defined $chosenWebs ) {
      $page .= CGI::hidden( -name => 'webs', -value => $chosenWebs );
    }
    if( defined $chosenUsers ) {
      $page .= CGI::hidden( -name => 'users', -value => $chosenUsers );
    }

    $page .= $tab . CGI::end_form();
    return $page;
}

sub TOPICPERMISSIONS {
    my( $session, $params, $topic, $web ) = @_;

    #this is to redirect to the "no access" page if this tag is used in a non-view template.
#    TWiki::UI::checkAccess( $session, $web, $topic,
#                                'view', $session->{user} );
	unless( &TWiki::Func::checkAccessPermission( 'VIEW', 
                    $session->{user}, undef, $topic, $web )) {
        throw TWiki::OopsException( 'accessdenied',
                                    def => 'topic_access',
                                    web => $web,
                                    topic => $topic,
                                    params =>
                                      [ 'VIEW',
                                        $session->security->getReason()]);
    }
   my $disableSave = 'Disabled';
   $disableSave = '' if TWiki::Func::checkAccessPermission( 'CHANGE', 
                    $session->{user}, undef, $topic, $web );

   my $pluginPubUrl = TWiki::Func::getPubUrlPath().'/'.
            TWiki::Func::getTwikiWebname().'/WebPermissionsPlugin';

    #add the JavaScript
    my $jscript = TWiki::Func::readTemplate ( 'webpermissionsplugin', 'topicjavascript' );
    $jscript =~ s/%PLUGINPUBURL%/$pluginPubUrl/g;
    TWiki::Func::addToHEAD('WebPermissionsPlugin', $jscript);

    my $templateText = TWiki::Func::readTemplate ( 'webpermissionsplugin', 'topichtml' );
    $templateText =~ s/%SCRIPT%/%SCRIPTURL{save}%/g if ($disableSave eq '');
    $templateText =~ s/%SCRIPT%/%SCRIPTURL{view}%/g unless ($disableSave eq '');
    $templateText = TWiki::Func::expandCommonVariables( $templateText, $topic, $web );

    my $topicViewerGroups = '';
    my $topicViewers = '';
    my $topicEditorGroups = '';
    my $topicEditors = '';
    my $unselectedGroups = '';
    my $unselectedUsers = '';

    my ( $acls, $inheritedPermissions ) = _getACLs( [ 'VIEW', 'CHANGE' ], $web, $topic);
    foreach my $user ( sort (keys %$acls) ) {
        my $isGroup;
        if (defined &TWiki::Func::isGroup) {
            $isGroup = TWiki::Func::isGroup( $user );
        } else {
            $isGroup = ($user =~ /Group$/);
        }
        if ( $acls->{$user}->{CHANGE} ) {
            $topicEditors .= '<OPTION>'.$user.'</OPTION>'
              unless $isGroup;
            $topicEditorGroups .= '<OPTION>'.$user.'</OPTION>'
              if $isGroup;
        } elsif ( $acls->{$user}->{VIEW} ) {
            $topicViewers .= '<OPTION>'.$user.'</OPTION>'
              unless $isGroup;
            $topicViewerGroups .= '<OPTION>'.$user.'</OPTION>'
              if $isGroup;
        } else {
            $unselectedUsers .= '<OPTION>'.$user.'</OPTION>'
              unless $isGroup;
            $unselectedGroups .= '<OPTION>'.$user.'</OPTION>'
              if $isGroup;
        }
    }
    $templateText =~ s/%EDITGROUPS%/$topicEditorGroups/g;
    $templateText =~ s/%EDITUSERS%/$topicEditors/g;
    $templateText =~ s/%VIEWGROUPS%/$topicViewerGroups/g;
    $templateText =~ s/%VIEWUSERS%/$topicViewers/g;
    $templateText =~ s/%UNSELECTEDGROUPS%/$unselectedGroups/g;
    $templateText =~ s/%UNSELECTEDUSERS%/$unselectedUsers/g;
    $templateText =~ s/%PLUGINNAME%/WebPermissionsPlugin/g;
    $templateText =~ s/%DISABLESAVE%/$disableSave/g;
	
	if ($inheritedPermissions) {
		$templateText =~ s/%CUSTOMCHECKED%//g;
		$templateText =~ s/%INHERITEDCHECKED%/checked/g;
	} else {
		$templateText =~ s/%CUSTOMCHECKED%/checked/g;
		$templateText =~ s/%INHERITEDCHECKED%//g;
	}

	my( $meta, $text ) = TWiki::Func::readTopic( $web, $topic );
	my $parent = $meta->getParent();
	if ($parent) {
		my ($parentWeb, $parentTopic) = TWiki::Func::normalizeWebTopicName( $web, $parent );
		if (&TWiki::Func::checkAccessPermission( 'VIEW', 
                    $session->{user}, undef, $parentTopic, $parentWeb )) {
			my $t = TWiki::Func::expandTemplate( 'copyoption' );
			$t =~ s/%COPYWEB%/$parentWeb/g;
			$t =~ s/%COPYTOPIC%/$parentTopic/g;
			$templateText =~ s/%COPYOPTION%/$t/g;
		} else {
			$templateText =~ s/%COPYOPTION%//g;
		}
	} else {
		$templateText =~ s/%COPYOPTION%//g;
	}
	
	my $children = _getChangeableChildren($web, $topic, $session);
	if (scalar(@$children)) {
		my $list = '';
		my $t = TWiki::Func::expandTemplate( 'applytochildren' );
		my $ct = TWiki::Func::expandTemplate( 'checkboxchildren' );
		foreach my $child (@$children) {
			$list .= $ct;
			$list =~ s/%CHILDTOPIC%/$child/g;
		}
		$t =~ s/%CHILDREN%/$list/g;
		$templateText =~ s/%APPLYTOCHILDREN%/$t/o;
	} else {
		$templateText =~ s/%APPLYTOCHILDREN%//g;
	}
				
    return $templateText;
}

sub beforeSaveHandler {
    my ( $text, $topic, $web, $meta, $query, $action ) = @_;

    if ($action ne 'Save') {
        #SMELL: canceling out from, or just stoping a save seems to be quite difficult
        TWiki::Func::redirectCgiQuery( $query, &TWiki::Func::getViewUrl( $web, $topic ) );
        throw Error::Simple( 'cancel permissions action' );
    }

    return if ($TWiki::Plugins::WebPermissionsPlugin::antiBeforeSaveRecursion == 1);
    $TWiki::Plugins::WebPermissionsPlugin::antiBeforeSaveRecursion = 1;

	my $acl_option = $query->param('acl_option');
	return unless $acl_option;
	
	my $applyToChildren = $query->param('applytochildren') || '';
	my $dontApplyToMe = $query->param('dontapplytome') || '';	
	
	my $context = 'TOPIC';
	my $settings;
	my @modes = split(/[\s,]+/,$TWiki::cfg{Plugins}{WebPermissionsPlugin}{modes} ||
							   'VIEW,CHANGE' );

	switch ($acl_option) {							   
		case('custom') {
			#these lists only contain selected users (by using javascript to select the changed ones in save onclick)
			my @topicEditors = $query->param('topiceditors');
			my @topicViewers = $query->param('topicviewers');
			my @disallowedUsers = $query->param('disallowedusers');

			if ((@topicEditors || @topicViewers || @disallowedUsers)) {
				#TODO: change this to get modes from params

				my ( $acls, $inheritedPermissions ) = _getACLs( \@modes, $web, $topic);
				my ($userName, $userObj);
				foreach $userName (@topicEditors) {
					$acls->{$userName}->{'CHANGE'} = 1;
					$acls->{$userName}->{'VIEW'} = 1;
				}
				foreach $userName (@topicViewers) {
					$acls->{$userName}->{'CHANGE'} = 0;
					$acls->{$userName}->{'VIEW'} = 1;
				}
				foreach $userName (@disallowedUsers) {
					$acls->{$userName}->{'CHANGE'} = 0;
					$acls->{$userName}->{'VIEW'} = 0;
				}

				#TODO: what exactly happens on error?
				$settings = _parseCustomACLs( \@modes, $acls, $context );
			} else {
				return;
			}
		}
		case('public') {
			$settings = _parseOptionACLs( \@modes, $context, 'DENY', '' );
		}
		case('group') {
			my $group = TWiki::Func::getPreferencesValue('ALLOWWEBVIEW',$web);
			$settings = _parseOptionACLs( \@modes, $context, 'ALLOW', $group );
		}
		case('private') {
			my $wikiname = TWiki::Func::getWikiName();
			$settings = _parseOptionACLs( \@modes, $context, 'ALLOW', $wikiname );
		}
		case('inherit') {
			$settings = {};
		}
		case('copy') {
			my( $meta, $text ) = TWiki::Func::readTopic( $web, $topic );
			my $parent = $meta->getParent();
			return unless $parent;
			my ($parentWeb, $parentTopic) = TWiki::Func::normalizeWebTopicName( $web, $parent );
			$settings = _copyACLs( \@modes, $parentWeb, $parentTopic, $context);
		}
	}

	my @topicList;
	if ($applyToChildren) {
		foreach my $child ( $query->param( 'childtopic' ) ) {
			if (TWiki::Func::topicExists($web, $child) && 
					TWiki::Func::checkAccessPermission( 'CHANGE', $TWiki::Plugins::SESSION->{user}, undef, $child, $web )) {
				$child = TWiki::Sandbox::untaintUnchecked($child);
				push (@topicList, $child);
			}
		}
		unshift (@topicList, $topic) unless $dontApplyToMe;
	} else {
		push (@topicList, $topic);
	}
	foreach (@topicList) {
		_writeACLs( \@modes, $settings, $web, $_, $context );
		#TWiki::Func::writeDebug("writing $_");
	}
	#read in what writeACLs just saved, (don't grok why redirect looses the save)
	($_[3], $_[0]) = TWiki::Func::readTopic($_[2],$_[1]);

	#SMELL: canceling out from, or just stoping a save seems to be quite difficult
	#return a redirect to view..
	TWiki::Func::redirectCgiQuery( $query, &TWiki::Func::getViewUrl( $web, $topic ) );
	throw Error::Simple( 'permissions action saved' );
}

# Filter a list of strings based on the filter expression passed in
sub _filterList {
    my $filter = shift;
    my %included;

    foreach my $expr (split(/,/, $filter)) {
        my $exclude = ($expr =~ s/^-//) ? 1 : 0;
        $expr =~ s/\*/.*/g;
        $expr =~ s/\?/./g;
		#&TWiki::Func::writeDebug($expr);
        foreach my $item (@_) {
            # The \s's are needed to retain compatibility
            if ($item =~ /^\s*$expr\s*$/) {
				if ($exclude) {
                    delete $included{$item};
                } else {
					#&TWiki::Func::writeDebug($item);
					$included{$item} = 1;
                }
            }
        }
    }
    return keys %included;
}

sub USERSLIST {
    my( $this, $params ) = @_;
    my $format = $params->{_DEFAULT} || $params->{'format'} || '$wikiname';
    my $separator = $params->{separator} || "\n";
    $separator =~ s/\$n/\n/;
    my $selection = $params->{selection} || '';
    $selection =~ s/\,/ /g;
    $selection = " $selection ";
    my $marker = $params->{marker} || 'selected="selected"';

    my @items;
    foreach my $item ( _getListOfUsers() ) {
        my $line = $format;
        $line =~ s/\$wikiname\b/$item/ge;
        my $mark = ( $selection =~ / \Q$item\E / ) ? $marker : '';
        $line =~ s/\$marker/$mark/g;
        if (defined(&TWiki::Func::decodeFormatTokens)) {
            $line = TWiki::Func::decodeFormatTokens( $line );
        } else {
            $line =~ s/\$n\(\)/\n/gs;
            $line =~ s/\$n([^$TWiki::regex{mixedAlpha}]|$)/\n$1/gs;
            $line =~ s/\$nop(\(\))?//gs;
            $line =~ s/\$quot(\(\))?/\"/gs;
            $line =~ s/\$percnt(\(\))?/\%/gs;
            $line =~ s/\$dollar(\(\))?/\$/gs;
        }
        push( @items, $line );
    }
    return join( $separator, @items);
}

# Get a list of all registered users
sub _getListOfUsers {
    my @list;
    if (defined(&TWiki::Func::eachUser)) {
        my $it = TWiki::Func::eachUser();
        while ($it->hasNext()) {
            my $user = $it->next();
            push(@list, $user);
        }
    } else {
        # Compatibility; pre 4.2
        my $session = $TWiki::Plugins::SESSION;
        my $users = $session->{users};

        #if we have the UserMapping changes (post 4.0.2)
        if (defined (&TWiki::Users::getAllUsers)) {
            @list = @{$users->getAllUsers()};
        } else {
            $users->lookupLoginName('guest'); # load the cache

            @list =
              map {
                  my( $w, $t ) = TWiki::Func::normalizeWebTopicName(
                      $TWiki::cfg{UsersWebName}, $_);
                  $users->findUser( $t, "$w.$t");
              } values %{$users->{U2W}};
        }
        @list = map($_->wikiName(), grep(!$_->isGroup(), grep($_, @list)));

    }
    return @list;
}

# Get a list of all groups
sub _getListOfGroups {
    my @list;
    if (defined(&TWiki::Func::eachGroup)) {
        my $it = TWiki::Func::eachGroup();
        while ($it->hasNext()) {
            my $user = $it->next();
            push(@list, $user);
        }
    } else {
        # Compatibility; pre 4.2
        my $session = $TWiki::Plugins::SESSION;
        my $users = $session->{users};

        # if we have the UserMapping changes (post 4.0.2)
        if (defined (&TWiki::Users::getAllGroups)) {
            @list = map { $_->wikiName() }
              @{$session->{users}->getAllGroups()};
        } else {
            # This code assumes we are using TWiki topic based Group mapping
            $session->{search}->searchWeb(
                _callback     => sub {
                    my $ref = shift;
                    my $group = shift;
                    return unless $group;
                    my $groupObject = $ref->{users}->findUser( $group );
                    push (@{$ref->{list}}, $groupObject->wikiName())
                      if $groupObject;
                },
                _cbdata       =>  { list => \@list, users => $users },
                inline        => 1,
                search        => "Set GROUP =",
                web           => 'all',
                topic         => "*Group",
                type          => 'regex',
                nosummary     => 'on',
                nosearch      => 'on',
                noheader      => 'on',
                nototal       => 'on',
                noempty       => 'on',
                format	     => "\$web.\$topic",
                separator     => '',
               );
        }
    }
    return @list;
}

sub _getACLs {
    my( $modes, $web, $topic ) = @_;

    my $context = 'TOPIC';
    unless( $topic ) {
        $context = 'WEB';
        $topic = $TWiki::cfg{WebPrefsTopicName};
    }

    my @knownusers = _getListOfUsers();
    push(@knownusers, _getListOfGroups());

    my %acls;

    # By default, allow all to access all
    foreach my $user ( @knownusers ) {
        foreach my $mode ( @$modes ) {
            $acls{$user}->{$mode} = 1;
        }
    }

    #print STDERR "Got users ",join(',',keys %acls),"\n";
    foreach my $mode ( @$modes ) {
        foreach my $perm ( 'ALLOW', 'DENY' ) {
            my $users = $TWiki::Plugins::SESSION->{prefs}->getWebPreferencesValue(
                    $perm.'WEB'.$mode, $web, $topic );
				&TWiki::Func::writeDebug($perm.'WEB'.$mode.' '.$users) if DEBUG;
			# if ALLOWWEBCHANGE is not defined, get the settings for ALLOWWEBVIEW
			if (($perm.'WEB'.$mode eq 'ALLOWWEBCHANGE') && !$users) {
				$users = $TWiki::Plugins::SESSION->{prefs}->getWebPreferencesValue(
                    'ALLOWWEBVIEW', $web, $topic );
				&TWiki::Func::writeDebug($perm.'WEBCHANGE'.' '.$users) if DEBUG;
			}
			next unless defined $users;
			if ($users =~ /\S$/) { 
				my $acls = _setExistingACLs($mode,$perm,\%acls,$users,\@knownusers);
				%acls = %$acls;
			}
			undef $users;
        }
    }
	
	my $inheritedPermissions = 1; # default, assume topic inherits its access control from its web
	if ($context eq 'TOPIC') {
		foreach my $mode ( @$modes ) {
			foreach my $perm ( 'ALLOW', 'DENY' ) {
				my $users = $TWiki::Plugins::SESSION->{prefs}->getTopicPreferencesValue(
                    $perm.'TOPIC'.$mode, $web, $topic );
                &TWiki::Func::writeDebug($perm.$context.$mode.' '.$users) if DEBUG;
				next unless defined $users;
				$inheritedPermissions = 0; # there's some sort of Access Control declaration in the topic
				if ($users =~ /\S$/ ) {
					my $acls = _setExistingACLs($mode,$perm,\%acls,$users,\@knownusers);
					%acls = %$acls;
				} else { # empty permissions list
					&TWiki::Func::writeDebug($perm.$context.$mode.' empty user list') if DEBUG;
					# empty DENY list = allow all (deprecated but still supported); empty ALLOW list means nothing
					if ($perm eq 'DENY') {
						foreach my $user ( @knownusers ) {
							$acls{$user}->{$mode} = 1;
						}
					}
				}
            }
		}
	}
	
	if (0) {
	 foreach my $myuser (@knownusers) {
		foreach my $mymode ('VIEW','CHANGE') {
			&TWiki::Func::writeDebug($myuser.' '.$mymode.' '.$acls{$myuser}->{$mymode});
		}
	 }
	}
    return ( \%acls, $inheritedPermissions );
}

sub _setExistingACLs {
	my( $mode, $perm, $acls, $users, $knownusers ) = @_;
	my %acls = %$acls;
            my @users =    # @lusers =
              grep { $_ }
                map {
                    my( $w, $t ) = TWiki::Func::normalizeWebTopicName(
                        $TWiki::cfg{UsersWebName}, $_);
                    $t;
                } split( /[ ,]+/, $users || '' );

# commented out the following because it introduces functionality that we don't (and shouldn't) support:
# TheAGroup is expanded that has CHANGE rights, the editor decides that he doesn't want member X to have any access at all
# so he moves him out of the Editor list and thinks, "Job done," but TheAGroup is still in the Editor list and, therefore, so is member X
# (there is no code that adds a lone DENYTOPICxxxx setting for an individual user that has been specifically removed from an ALLOWTOPIC list).
# The correct way to do this is to create a new group without member X and have TheAGroup made up of the new group plus member X.
            # expand groups
            # my @users;
            # while( scalar( @lusers )) {
                # my $user = pop( @lusers );
                # my $isGroup = TWiki::Func::isGroup( $user );
                # if( $isGroup) {
                    # expand groups and add individual users
                    # my $it = TWiki::Func::eachGroupMember($user);
                    # while ($it && $it->hasNext()) {
                        # push( @lusers, $it->next() );
                    # }
                # }
                # push( @users, $user );
            # }

            if( $perm eq 'ALLOW' ) {
                # If ALLOW, only users in the ALLOW list are permitted,
                # so change the default for all other users to 0.
                foreach my $user ( @{$knownusers} ) {
                    #print STDERR "Disallow ",$user,"\n";
					&TWiki::Func::writeDebug('Disallow'.$mode.' '.$user) if DEBUG;
                    $acls{$user}->{$mode} = 0;
                }
                foreach my $user ( @users ) {
                    #print STDERR "Allow ",$user,"\n";
					&TWiki::Func::writeDebug('ALLOW'.$mode.' '.$user) if DEBUG;
                    $acls{$user}->{$mode} = 1;
                }
            } else {
                foreach my $user ( @users ) {
					&TWiki::Func::writeDebug('DENY'.$mode.' '.$user) if DEBUG;
                    #print STDERR "Deny ",$user,"\n";
                    $acls{$user}->{$mode} = 0;
                }
            }
	return \%acls;
}

sub _parseCustomACLs {
    my( $modes, $acls, $context ) = @_;

    my @knownusers = _getListOfUsers();
    push(@knownusers, _getListOfGroups() );

	my $settings = {};
    foreach my $op ( @$modes ) {
        my @allowed = grep { $acls->{$_}->{$op} } @knownusers;
        my @denied = grep { !$acls->{$_}->{$op} } @knownusers;

        if( scalar( @denied )) {
            # Work out the access modes
            my $name;
            my $set;
            if( scalar( @denied ) <= scalar( @allowed )) {
                $name = 'DENY'.$context.$op;
                $set = \@denied;
            } elsif (scalar( @allowed )) { # an empty allow list means nothing - better to have no setting at all
                $name = 'ALLOW'.$context.$op;
                $set = \@allowed;
            } else {
				next;
			}
			my @users = @$set;
			# TWiki::Func::writeDebug(join(' ',@users));
            while ( scalar(@$set) ) {
				my $user = pop( @$set );
                my $isGroup = TWiki::Func::isGroup( $user );
                if ($isGroup) {
                    # expand groups and remove their users from the @$set list
                    my $it = TWiki::Func::eachGroupMember($user);
                    while ($it && $it->hasNext()) {
						$user = $it->next();
						# TWiki::Func::writeDebug("deleting $next");
                        @users = grep(!/$user/, @users);
						# TWiki::Func::writeDebug(join(' ',@users));
                    }
                }
			}
			$settings->{$name} = join(' ', @users);
        } else { # no restrictions on the page but be explicit in the ACL declarations incase the parent Web has restrictions
			$settings->{'DENY'.$context.$op} = '';
		}
    }

	return $settings;
}

sub _parseOptionACLs {
    my( $modes, $context, $perm, $value ) = @_;

	my $settings = {};
    foreach my $op ( @$modes ) {
		my $name = $perm.$context.$op;
		$settings->{$name} = $value;
    }

	return $settings;
}

sub _removeOldACLs {
    my( $modes, $web, $topic, $context ) = @_;

	my( $meta, $text ) = TWiki::Func::readTopic( $web, $topic );

    foreach my $op ( @$modes ) {
        # Remove existing preferences of this type in text
        $text =~ s/^(   |\t)+\* Set (ALLOW|DENY)$context$op =.*$//gm;
        $meta->remove('PREFERENCE', 'DENY'.$context.$op);
        $meta->remove('PREFERENCE', 'ALLOW'.$context.$op);
	}
	
	return ( $meta, $text );
}

sub _copyACLs {
	my ($modes, $web, $topic, $context) = @_;
	
	my( $meta, $text ) = TWiki::Func::readTopic( $web, $topic );
	
	my %settings;
    foreach my $op ( @$modes ) {
		my $deny = 'DENY'.$context.$op;
		my $allow = 'ALLOW'.$context.$op;
		foreach my $setting ( $deny, $allow ) {
	        if ($text =~ m/^(?:   |\t)+\* Set $setting =(.*)$/) {
				$settings{$setting} = $1;
			}
	        my $ref = $meta->get('PREFERENCE', $setting);
			if (defined $ref) {
				$settings{$setting} = $ref->{'value'};
			}
		}
	}

	return \%settings;
}

sub _writeACLs {
	my ($modes, $settings, $web, $topic, $context) = @_;

	my( $meta, $text ) = _removeOldACLs($modes, $web, $topic, $context);
	
    if( $TWiki::cfg{Plugins}{WebPermissionsPlugin}{UsePlainText} ) {
        $text .= "\n" unless $text =~ /\n$/s;
    }

	foreach my $setting (keys %$settings) {
		if ($TWiki::cfg{Plugins}{WebPermissionsPlugin}{UsePlainText}) {
			$text .= "   * Set $setting = ".$settings->{$setting}."\n";
		} else {
			$meta->putKeyed( 'PREFERENCE',
							 {
								 name => $setting,
								 type => 'Set',
								 title => $setting,
								 value => $settings->{$setting}
								}
							);
		}
	}

    # If there is an access control violation this will throw.
    TWiki::Func::saveTopic( $web, $topic, $meta, $text, { minor => 1 } );
}

sub _getChangeableChildren {
	my ($web, $topic, $session) = @_;
	my $children = _getChildren($web, $topic);
	my @changeableKids;
	foreach my $child (@$children) {
		if (TWiki::Func::checkAccessPermission( 'CHANGE', $session->{user}, undef, $child, $web )) {
				push(@changeableKids, $child);
		}
	}
	return \@changeableKids;
}

# adapted the following two from GenPDF

sub _getChildren {
	my ($web, $topic) = @_;
	
	my $search="%SEARCH{\"parent != \'\'\" web=\"$web\" format=\"\$parent \$topic\" type=\"query\" nonoise=\"on\"}%";
    #&TWiki::Func::writeDebug($search);    
	my %tree;
    my $result = TWiki::Func::expandCommonVariables($search);
	#&TWiki::Func::writeDebug($result);
	foreach ( split /\n/, $result ) {
		$_ =~ /([A-Za-z0-9\-_]+)\s([A-Za-z0-9\-_]+)/ && do {
			#&TWiki::Func::writeDebug("$1: $2");
			push @{ $tree{$1} }, $2;
			#push @{ $tree{$parent} }, $child;
		}
	}
	# Do a recursive depth first walk through the ancestors in the tree
	my @topics;
	_depthFirst($topic, \@topics, \%tree);
	#TWiki::Func::writeDebug( join(' ', @topics) );
	return \@topics;
}

sub _depthFirst {
    my ($parent, $topics, $tree) = @_;
	my %tree = %$tree;
    # the grep gets around a perl dereferencing bug when using strict refs
    my @children = grep(/.*/, @{$tree{$parent}});
	#TWiki::Func::writeDebug( join(' ', @children) );
    for ( sort @children ) {
        #TWiki::Func::writeDebug( "new child of $parent: '$_'\n"); # DEBUG
        push @$topics, $_;
        if (defined $tree{$_}) {
			# this child is also a parent so bring them in too
			_depthFirst($_, $topics, $tree);
        }
    }
}


1;
